﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.IO.Pipes;
using System.Windows.Forms;
using System.Threading;
using System.Runtime.InteropServices;

namespace EditorTestbed
{
    public class Client
    {
        enum ConnectionState
        {
            WAITING,
            ACTIVE,
            CLOSED,
            FAILED
        };

        public Client(Testbed app)
        {
            m_app = app;
        }

        public bool Connect()
        {
            pipeWrite =
                new NamedPipeClientStream("IPCTestPipe_server_read");
            pipeRead =
                new NamedPipeClientStream("localhost", 
                    "IPCTestPipe_server_write", PipeDirection.In);
                    //PipeOptions.Asynchronous);

            m_state = ConnectionState.WAITING;
            try
            {
                int timeOut = 1000;
                pipeRead.Connect(timeOut);
                pipeWrite.Connect(timeOut);
            }
            catch (Exception)
            {
                m_state = ConnectionState.CLOSED;
                MessageBox.Show("Cannot connect to server, is it running?");
                return false;
            }

            m_app.ConsolePrintln("Connected to server.");
            m_state = ConnectionState.ACTIVE;
            m_writeThread = new Thread(this.WriteThread);
            m_writeThread.Start();
            m_readThread = new Thread(this.ReadThread);
            m_readThread.Start();

            //pipeRead.BeginRead(m_headerBuffer, 0, m_headerBuffer.Length, 

            return true;
        }

        public void Close()
        {
            if (pipeWrite == null || pipeRead == null)
                return;

            m_terminating = true;
            pipeWrite.Close();
            pipeRead.Close();
            m_writeQueue.Signal();
            m_readQueue.Signal();
            // Read thread may be stuck at Pipe::Read().
            // I couldnt find a more elegant way to break it (without bothering
            // with async reading), so I interrupt the thread.
            try
            {
                m_writeThread.Join();
                m_readThread.Interrupt();
            }
            catch (Exception)
            {
            }
            m_writeThread = null;
            m_readThread = null;
            m_state = ConnectionState.CLOSED;
            m_app.OnDisconnect();
            m_terminating = false;
        }

        static byte[] StructureToByteArray(Object obj)
        {
            int numBytes = Marshal.SizeOf(obj);
            byte[] rawArray = new byte[numBytes];
            IntPtr ptr = Marshal.AllocHGlobal(numBytes);
            Marshal.StructureToPtr(obj, ptr, true);
            Marshal.Copy(ptr, rawArray, 0, numBytes);
            Marshal.FreeHGlobal(ptr);
            return rawArray;
        }
        static void StructureToByteArray(Object obj, byte[] rawArray, int offset)
        {
            int numBytes = Marshal.SizeOf(obj);
            IntPtr ptr = Marshal.AllocHGlobal(numBytes);
            Marshal.StructureToPtr(obj, ptr, true);
            Marshal.Copy(ptr, rawArray, offset, numBytes);
            Marshal.FreeHGlobal(ptr);
        }
        public static Object ByteArrayToStructure(byte[] arr, Type objType)
        {
            System.Diagnostics.Debug.Assert(Marshal.SizeOf(objType) == arr.Length);
            IntPtr sptr = Marshal.AllocHGlobal(arr.Length);
            Marshal.Copy(arr, 0, sptr, arr.Length);
            Object obj = Marshal.PtrToStructure(sptr, objType);
            Marshal.FreeHGlobal(sptr);
            return obj;
        }
        public Message CreateMessage(MessageID id, uint size, MessageType type)
        {
            Message msg = new Message();
            msg.header.id = (uint)id;
            msg.header.size = size;
            msg.header.transaction = m_nextTransaction++;
            msg.header.type = (uint)type;
            if (size > 0)
                msg.data = new byte[size];
            return msg;
        }
        public Message CreateMessageWithData(MessageID id, MessageType type, Object data)
        {
            Message msg = CreateMessage(id, (uint)Marshal.SizeOf(data), type);
            StructureToByteArray(data, msg.data, 0);
            return msg;
        }

        void WriteThread()
        {
            while (!m_terminating && m_state != ConnectionState.CLOSED)
            {
                m_writeQueue.WaitForItem();
                Message msg = m_writeQueue.Dequeue();
                if (msg != null && !m_terminating)
                {
                    WriteMessage(msg);
                }
            }
        }
        void ReadThread()
        {
            while (!m_terminating && m_state != ConnectionState.CLOSED)
            {
                if (!ReadPump())
                    break;
            }
        }
        bool ReadPump()
        {
            Message msg = ReadMessage();
            if (msg != null)
            {
                if (msg.header.type == (uint)MessageType.PROTOCOL)
                    HandleProtocolMessage(msg);
                else
                {
                    System.Diagnostics.Debug.Assert(msg.header.type == (uint)MessageType.USER);
                    m_readQueue.Enqueue(msg);
                }
                return true;
            }
            else
            {
                Close();
                return false;
            }
        }
        void HandleProtocolMessage(Message msg)
        {
            switch (msg.header.id)
            {
                case 0:
                    // Disconnect
                    Close();
                    break;
                default:
                    System.Diagnostics.Debug.Fail("Unknown protocol message received");
                    break;
            }
        }

        Message ReadMessage()
        {
            byte[] header = new byte[16];
            int bytesRead = pipeRead.Read(header, 0, header.Length);
            if (bytesRead == header.Length)
            {
                Message msg = new Message();
                msg.header = (MessageHeader)ByteArrayToStructure(header, typeof(MessageHeader));
                if (msg.header.size > 0)
                {
                    msg.data = new byte[msg.header.size];
                    pipeRead.Read(msg.data, 0, (int)msg.header.size);
                }
                return msg;
            }
            return null;
        }

        void WriteMessage(Message msg)
        {
            try
            {
                byte[] headerData = StructureToByteArray(msg.header);
                pipeWrite.Write(headerData, 0, headerData.Length);
                if (msg.header.size > 0 && !m_terminating)
                    pipeWrite.Write(msg.data, 0, msg.data.Length);
            }
            catch (System.IO.IOException)
            {
            }
        }

        public bool Send(Message msg)
        {
            m_stampedMessages.AddMessage(msg);
            m_writeQueue.Enqueue(msg);
            return true;
        }
        public Message Receive()
        {
            return m_readQueue.Dequeue();
        }

        ConnectionState m_state = ConnectionState.CLOSED;
        bool m_terminating = false;
        int m_nextTransaction = 1;
        NamedPipeClientStream pipeRead;
        NamedPipeClientStream pipeWrite;
        MessageQueue m_writeQueue = new MessageQueue();
        MessageQueue m_readQueue = new MessageQueue();
        public StampedMessages m_stampedMessages = new StampedMessages();
        Thread m_writeThread;
        Thread m_readThread;
        byte[] m_headerBuffer = new byte[16];
        Testbed m_app;
    }
}
