﻿#define SEND_MOUSE_MOVE 
// Mouse movement can be either send by pipe from the client or
// retrieved using GetCursorPos(). The latter is obviously more
// effective, but it makes recording messages for later replay
// impossible.

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.IO;
using WeifenLuo.WinFormsUI.Docking;


namespace EditorTestbed
{
    public partial class Testbed : Form
    {
        public static Testbed ms_instance = null;

        public Testbed()
        {
            InitializeComponent();
            m_deserializeDockContent = new DeserializeDockContent(GetContentFromPersistString);
            ms_instance = this;
        }

        private IDockContent GetContentFromPersistString(string persistString)
        {
            if (persistString == typeof(MessageStack).ToString())
                return m_messageStack;
            else if (persistString == typeof(OutputWindow).ToString())
                return m_outputWindow;
            else if (persistString == typeof(RenderArea).ToString())
                return m_renderArea;
            else if (persistString == typeof(ObjectProperties).ToString())
                return m_objectProperties;
            return null;
        }

        private MessageStack m_messageStack;
        private OutputWindow m_outputWindow = new OutputWindow();
        private ObjectProperties m_objectProperties = new ObjectProperties();
        private DeserializeDockContent m_deserializeDockContent;
        private Client m_client;
        private RenderArea m_renderArea = new RenderArea();
        public World m_world = new World();
        private Entity m_selectedEntity = null;
        private bool m_connected = false;

        private void Testbed_Shown(object sender, EventArgs e)
        {
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        private void Testbed_FormClosing(object sender, FormClosingEventArgs e)
        {
            m_client.Close();

            string configFile = Path.Combine(Path.GetDirectoryName(Application.ExecutablePath), "DockPanel.config");
            dockPanel.SaveAsXml(configFile);
        }

        private void Testbed_Load(object sender, EventArgs e)
        {
            m_client = new Client(this);
            m_messageStack = new MessageStack(m_client);
            string configFile = Path.Combine(Path.GetDirectoryName(Application.ExecutablePath), "DockPanel.config");
            if (File.Exists(configFile))
                dockPanel.LoadFromXml(configFile, m_deserializeDockContent);
        }

        private void toolStripButtonConnect_Click(object sender, EventArgs e)
        {
            if (m_client.Connect())
            {
                objectsToolStripMenuItem.Enabled = true;
                toolStripButtonConnect.Enabled = false;

                Message_Handshake hwnd;
                hwnd.hwnd = m_renderArea.GetRenderHWND();
                hwnd.width = m_renderArea.GetRenderWidth();
                hwnd.height = m_renderArea.GetRenderHeight();
                hwnd.nextID = m_world.GetNextID();
                ConsolePrintln("Handshaking, render window set to 0x" + hwnd.hwnd.ToString("X"));
                Message msg = m_client.CreateMessageWithData(MessageID.HANDSHAKE, (uint)MessageType.USER, hwnd);
                m_client.Send(msg);
                mainTimer.Start();
            }
        }

        void SendWorldState()
        {
            if (!m_connected)
                return;
            foreach (Entity ent in m_world.Entities)
            {
                Message_CreateObject msgData;
                msgData.objectType = (byte)ent.m_objectType;
                msgData.id = ent.m_id;
                Message msg = m_client.CreateMessageWithData(MessageID.CREATE_OBJECT, MessageType.USER, msgData);
                m_client.Send(msg);
                SendObjectProperties(ent);
            }
        }

        delegate void EnableButtonCallback(bool enabled);
        bool m_refreshRenderArea = false;
        public void OnDisconnect()
        {
            if (!m_connected)
                return;

            // Sometimes this seems to interrupt waiting thread(?)
            try
            {
                ConsolePrintln("Disconnected.");
            }
            catch (Exception)
            {
            }
            EnableButtonCallback d = new EnableButtonCallback(EnableConnectButton);
            Invoke(d, new object[] { true });
            m_connected = false;
            m_refreshRenderArea = true;
        }
        void EnableConnectButton(bool enabled)
        {
            toolStripButtonConnect.Enabled = enabled;
        }

        public void ConsolePrintln(string msg)
        {
            m_outputWindow.Println(msg);
        }

        void CreateMesh(ObjectType objType)
        {
            if (m_connected)
            {
                Message_CreateObject msgData;
                msgData.objectType = (byte)objType;
                msgData.id = 0xFFFFFFFF;    // Let server generate an ID.
                Message msg = m_client.CreateMessageWithData(MessageID.CREATE_OBJECT, MessageType.USER, msgData);
                m_client.Send(msg);
            }
        }

        private void createCubeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            CreateMesh(ObjectType.CUBE);
        }

        public void SendObjectProperties(Entity ent)
        {
            Message_ObjectProperties msgData;
            msgData.id = ent.m_id;
            msgData.color = (uint)ent.MeshColor.ToArgb();
            msgData.x = ent.m_position.x;
            msgData.y = ent.m_position.y;
            msgData.z = ent.m_position.z;
            msgData.vitalityPoints = ent.VitalityPoints;
            Message msg = m_client.CreateMessageWithData(MessageID.CHANGE_OBJECT_PROPERTIES, MessageType.USER, msgData);
            m_client.Send(msg);
        }

        private void toolStripButtonCrashGame_Click(object sender, EventArgs e)
        {
            Message msg = m_client.CreateMessage(MessageID.CRASH_GAME, 0, MessageType.USER);
            m_client.Send(msg);
        }

        void HandleMessage(Message msg)
        {
            MessageID id = (MessageID)msg.header.id;
            switch (id)
            {
                case MessageID.HANDSHAKE:
                    m_connected = true;
                    SendWorldState();
                    break;
                case MessageID.SELECT_OBJECT:
                    {
                        Message_ObjectID msgData = (Message_ObjectID)Client.ByteArrayToStructure(msg.data,
                            typeof(Message_ObjectID));
                        Entity ent = m_world.FindEntity(msgData.id);
                        m_selectedEntity = ent;
                        m_objectProperties.OnObjectSelected(ent);
                    }
                    break;

                case MessageID.CREATE_OBJECT:
                    {
                        Message_CreateObject msgData = (Message_CreateObject)Client.ByteArrayToStructure(msg.data,
                            typeof(Message_CreateObject));

                        System.Diagnostics.Debug.Assert(m_world.FindEntity(msgData.id) == null);
                        Entity ent = new Entity();
                        ent.m_objectType = (ObjectType)msgData.objectType;
                        ent.m_id = msgData.id;
                        m_world.AddEntity(ent);
                    }
                    break;

                case MessageID.CHANGE_OBJECT_PROPERTIES:
                    {
                        Message_ObjectProperties msgData = (Message_ObjectProperties)Client.ByteArrayToStructure(
                            msg.data, typeof(Message_ObjectProperties));
                        Entity ent = m_world.FindEntity(msgData.id);
                        
                        ent.m_position.x = msgData.x;
                        ent.m_position.y = msgData.y;
                        ent.m_position.z = msgData.z;
                        // Game cant change it.
                        //ent.MeshColor = Color.FromArgb((int)msgData.color);
                    }
                    break;

                case MessageID.DELETE_OBJECT:
                    {
                        Message_ObjectID msgData = (Message_ObjectID)Client.ByteArrayToStructure(msg.data,
                            typeof(Message_ObjectID));

                        Entity ent = m_world.FindEntity(msgData.id);
                        if (ent == m_selectedEntity)
                            m_selectedEntity = null;
                        m_world.RemoveEntity(ent);
                    }
                    break;

                default:
                    System.Diagnostics.Debug.Fail("Unsupported message ID: " + id.ToString());
                    break;
            }
        }

        void MessagePump()
        {
            while (true)
            {
                Message msg = m_client.Receive();
                if (msg == null)
                    break;
                HandleMessage(msg);
            }
        }

        private void mainTimer_Tick(object sender, EventArgs e)
        {
            MessagePump();
            if (m_refreshRenderArea)
            {
                m_renderArea.Refresh();
                m_refreshRenderArea = false;
            }
        }

        private void createSphereToolStripMenuItem_Click(object sender, EventArgs e)
        {
            CreateMesh(ObjectType.SPHERE);
        }


        public void OnObjectDelete()
        {
            if (m_selectedEntity != null && m_connected)
            {
                Message_ObjectID msgData;
                msgData.id = m_selectedEntity.m_id;
                Message msg = m_client.CreateMessageWithData(MessageID.DELETE_OBJECT, MessageType.USER, msgData);
                m_client.Send(msg);
            }
        }
        public void OnKeyboardEvent(Keys k, MessageID id)
        {
            if (m_connected)
            {
                Message_KeyboardEvent msgData;
                msgData.key = (int)k;
                Message msg = m_client.CreateMessageWithData(id, MessageType.USER, msgData);
                m_client.Send(msg);
            }
        }
        public void OnMouseEvent(MouseButton button, MessageID id)
        {
            if (m_connected)
            {
                Message_MouseEvent msgData;
                msgData.button = (byte)button;
                Message msg = m_client.CreateMessageWithData(id, MessageType.USER, msgData);
                m_client.Send(msg);
            }
        }

        public void OnMouseMoveEvent(int x, int y)
        {
#if SEND_MOUSE_MOVE
            if (m_connected && !m_messageStack.RealtimeReplay)
            {
                Message_MouseMoveEvent msgData;
                msgData.x = x;
                msgData.y = y;
                Message msg = m_client.CreateMessageWithData(MessageID.MOUSE_MOVE, MessageType.USER, msgData);
                m_client.Send(msg);
            }
#endif
        }

        public void OnViewportResize(int width, int height)
        {
            if (m_connected)
            {
                Message_Viewport msgData;
                msgData.width = width;
                msgData.height = height;
                Message msg = m_client.CreateMessageWithData(MessageID.VIEWPORT_RESIZE, MessageType.USER, msgData);
                m_client.Send(msg);
            }
        }

        private void createTorusToolStripMenuItem_Click(object sender, EventArgs e)
        {
            CreateMesh(ObjectType.TORUS);
        }

        private void saveSceneToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (saveWorldDialog.ShowDialog() == DialogResult.OK)
            {
                m_world.Save(saveWorldDialog.FileName);
            }
        }

        private void loadSceneToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (openWorldDialog.ShowDialog() == DialogResult.OK)
            {
                if (m_connected)
                {
                    Message msg = m_client.CreateMessage(MessageID.CLEAR_WORLD, 0, MessageType.USER);
                    m_client.Send(msg);
                }
                m_world.Load(openWorldDialog.FileName);
                SendWorldState();
            }
        }

        private void outputToolStripMenuItem_Click(object sender, EventArgs e)
        {
            m_outputWindow.Show(dockPanel);
        }

        private void objectPropertiesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            m_objectProperties.Show(dockPanel);
        }

        private void messagesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            m_messageStack.RefreshList(m_client.m_stampedMessages);
            m_messageStack.Show(dockPanel);
        }

        private void generateCHeaderToolStripMenuItem_Click(object sender, EventArgs e)
        {
            MessageReflection.ReflectTypes("Messages.h");
        }
    }
}

