#include "ipc/PipeConnection.h"
#include "ipc/Message.h"
#include "core/Console.h"
#include "core/LockGuard.h"
#include "core/RdeAssert.h"
#include <cstdio>
#include <cstring>

namespace rde
{
PipeConnection::PipeConnection()
:	m_terminating(false),
	m_state(ConnectionState::CLOSED)
{
	m_name[0] = '\0';
	m_pipeName[0] = '\0';
	m_readName[0] = '\0';
	m_writeName[0] = '\0';
	m_nextTransaction = -1;
}

PipeConnection::~PipeConnection()
{
	Close();
	m_readQueue.Clear();
	m_writeQueue.Clear();
}

bool PipeConnection::ServerInit(const char* name, const char* pipeName)
{
	strcpy_s(m_name, name);
	strcpy_s(m_pipeName, pipeName);
    sprintf_s(m_readName, "\\\\.\\pipe\\%s_server_read", m_pipeName);
    sprintf_s(m_writeName, "\\\\.\\pipe\\%s_server_write", m_pipeName);

	SetState(ConnectionState::WAITING);
	if (!m_connectThread.Start<PipeConnection, &PipeConnection::ServerThread>(this))
	{
		Console::Warningf("%s: Failed to create server thread\n", m_name);
		SetState(ConnectionState::FAILED);
		return false;
	}
	return true;
}

void PipeConnection::Close()
{
	if (!m_terminating)
	{
		m_terminating = true;
		m_terminateEvent.Signal();
		
		if (m_connectThread.IsRunning())
		{
			m_connectThread.Stop();
		}
		SetState(ConnectionState::CLOSED);

		m_terminating = false;
		m_terminateEvent.Reset();
	}
}

Message* PipeConnection::CreateMessage(uint32_t id, uint32_t size, uint32_t type)
{
	RDE_ASSERT(m_nextTransaction != 0);
	return Message::Create(id, m_nextTransaction--, size, type);
}

bool PipeConnection::Send(Message* msg)
{
	RDE_ASSERT(msg);
	if (m_state == ConnectionState::ACTIVE)
	{
		m_writeQueue.Enqueue(msg);
		return true;
	}
	return false;
}

bool PipeConnection::Receive(Message** msg, bool wait)
{
	if (m_state != ConnectionState::ACTIVE)
	{
		*msg = 0;
		return false;
	}
	if (wait)
		m_readQueue.WaitForItem();
	*msg = m_readQueue.Dequeue();
	return *msg != 0;
}

bool PipeConnection::SendProtocolMessage(ProtocolMessage::Enum msgID)
{
	Message* msg = CreateMessage(msgID, 0, MessageType::PROTOCOL);
	const bool ok = WriteMessage(msg);
	delete msg;
	return ok;
}

void PipeConnection::ServerThread()
{
	Thread::SetName("PipeServerThread");
	Console::Printf("%s: Starting pipe server '%s'\n", m_name, m_pipeName);
	while (!m_terminating)
	{
		if (!m_readPipe.Create(m_readName))
		{
			SetState(ConnectionState::FAILED);
			return;
		}
		if (!m_writePipe.Create(m_writeName))
		{
			SetState(ConnectionState::FAILED);
			return;
		}

		Console::Printf("%s: Ready for client [%s]\n", m_name, m_pipeName);

		while (!m_terminating)
		{
			const bool readPipeOk = m_readPipe.Connect(m_terminateEvent);
			const bool writePipeOk = m_writePipe.Connect(m_terminateEvent);
			if (readPipeOk && writePipeOk) // connected!
				break;
			else
				Thread::Sleep(200);
		}
		if (!m_terminating)
			ConnectThread();

		m_readPipe.Close();
		m_writePipe.Close();
		if (!m_terminating)
			SetState(ConnectionState::WAITING);
	} // while (!m_terminating)
	Console::Printf("%s: Stopping pipe server '%s'\n", m_name, m_pipeName);
}


void PipeConnection::ConnectThread()
{
	Console::Printf("%s: Client connected.\n", m_name);
	SetState(ConnectionState::ACTIVE);
	m_readThread.Start<PipeConnection, &PipeConnection::ReadThread>(this);
	m_writeThread.Start<PipeConnection, &PipeConnection::WriteThread>(this);
	// Wait for the read thread to finish.
	m_readThread.Stop();
	// Wake up readers.
	m_readQueue.Signal();
	m_writeQueue.Signal();
	m_writeThread.Stop();
	m_readQueue.Clear();
	m_writeQueue.Clear();

	SendProtocolMessage(ProtocolMessage::DISCONNECT);
}

void PipeConnection::ReadThread()
{
	Thread::SetName("PipeReadThread");
	while (!m_terminating && m_state != ConnectionState::CLOSED)
	{
		if (!ReadPump())
			break;
	}
}

bool PipeConnection::ReadPump()
{
	Message* msg(0);
	const bool readOk = ReadMessage(&msg);
	if (msg)
	{
		if (msg->GetHeader().type == MessageType::PROTOCOL)
		{
			HandleProtocolMessage(msg);
		}
		else //if (msg->GetHeader().type == MessageType::USER)
		{
			RDE_ASSERT(msg->GetHeader().type == MessageType::USER);
			m_readQueue.Enqueue(msg);
		}
#if RDE_IPC_DEBUG_LEVEL > 1
		const Message::Header& hdr = msg->GetHeader();
		Console::Printf("%s: Got message: id '%d', transaction '%d', size %d byte(s)\n",
			m_name, hdr.id, hdr.transaction, hdr.size);
		RDE_ASSERT(hdr.size >= 0);
#endif
	}
	if (!readOk)
		SetState(ConnectionState::CLOSED);
	return readOk;
}

bool PipeConnection::ReadMessage(Message** msg)
{
	Message::Header header;
	if (!Read(&header, sizeof(header)))
		return false;

	*msg = CreateMessage(header.id, header.size, header.type);
	if (header.size > 0 && !Read((*msg)->GetData(), header.size))
	{
		delete *msg;
		*msg = 0;
		return false;
	}
	return true;
}

void PipeConnection::HandleProtocolMessage(Message* msg)
{
	const Message::Header& hdr = msg->GetHeader();
	switch (hdr.id)
	{
	case ProtocolMessage::DISCONNECT:
		SetState(ConnectionState::CLOSED);
		break;
	default:
		RDE_ASSERT(!"Unknown protocol message ID");
	}
	delete msg;
}

void PipeConnection::WriteThread()
{
	Thread::SetName("PipeWriteThread");
	while (!m_terminating && m_state != ConnectionState::CLOSED)
	{
		if (!WritePump())
			break;
	}
}

bool PipeConnection::WritePump()
{
	bool result = true;
	m_writeQueue.WaitForItem();
	Message* msg = m_writeQueue.Dequeue();
	if (msg)
	{
		result = WriteMessage(msg);
#if RDE_IPC_DEBUG_LEVEL > 2
		const Message::Header& hdr = msg->GetHeader();
		Console::Printf("%s: Sent message '%d', transaction '%d', size '%d'\n",
			m_name, hdr.id, hdr.transaction, hdr.size);
#endif
		delete msg;
	}
	if (!result)
		SetState(ConnectionState::CLOSED);
	return result;
}

bool PipeConnection::WriteMessage(const Message* msg)
{
	const Message::Header& hdr = msg->GetHeader();
	if (!Write(&hdr, sizeof(hdr)))
		return false;
	return Write(msg->GetData(), hdr.size);
}

bool PipeConnection::Read(void* buffer, uint32_t bytes)
{
#if RDE_IPC_DEBUG_LEVEL > 2
	Console::Printf("%s: Trying to read %d byte(s)\n", m_name, bytes);
#endif
	RDE_ASSERT(bytes < 1 << 31);
	int32_t bytesLeft = (int32_t)bytes;
	while (bytesLeft > 0)
	{
		uint32_t bytesRead(0);
		const uint32_t bytesToRead = bytes < Pipe::kBufferSize ? bytes : Pipe::kBufferSize;
	
		if (!m_readPipe.Read(buffer, bytesToRead, m_terminateEvent, bytesRead))
			return false;
		if (m_terminating)
			return false;

		bytesLeft -= bytesRead;
#if RDE_IPC_DEBUG_LEVEL > 2
	Console::Printf("%s: Got %d byte(s), %d to go.\n", m_name, bytesRead, bytesLeft);
#endif
	}
	RDE_ASSERT(bytesLeft == 0);
#if RDE_IPC_DEBUG_LEVEL > 2
	Console::Printf("%s: Done, read %d byte(s).\n", m_name, bytes);
#endif
	return true;
}

bool PipeConnection::Write(const void* buffer, uint32_t bytes)
{
#if RDE_IPC_DEBUG_LEVEL > 2
	Console::Printf("%s: Trying to write %d byte(s)\n", m_name, bytes);
#endif
	RDE_ASSERT(bytes < 1 << 31);
	int32_t bytesLeft = (int32_t)bytes;
	while (bytesLeft > 0)
	{
		uint32_t bytesWritten(0);
		const uint32_t bytesToWrite = bytes < Pipe::kBufferSize ? bytes : Pipe::kBufferSize;
	
		if (!m_writePipe.Write(buffer, bytesToWrite, m_terminateEvent, bytesWritten))
			return false;
		if (m_terminating)
			return false;

		bytesLeft -= bytesWritten;
#if RDE_IPC_DEBUG_LEVEL > 2
		Console::Printf("%s: Written %d byte(s), %d to go.\n", m_name, bytesWritten, bytesLeft);
#endif
	}
	RDE_ASSERT(bytesLeft == 0);
#if RDE_IPC_DEBUG_LEVEL > 2
	Console::Printf("%s: Done, written %d byte(s).\n", m_name, bytes);
#endif
	return true;
}

void PipeConnection::SetState(ConnectionState::Enum state)
{
	LockGuard<Mutex> lock(m_stateMutex);
	if (m_state != state)
	{
		static const char* ls_stateNames[] =
		{
			"Waiting", "Active", "Closed", "Failed"
		};

		Console::Printf("%s: State changing from '%s' to '%s'\n", m_name,
			ls_stateNames[m_state], ls_stateNames[state]);
		m_state = state;
	}
}

}
