#include "PreCompiled.h"
#include "Scene.h"
#include "Entity.h"
#include "Mesh.h"
#include "DX9Utils.h"

Scene::Scene(const rde::ComPtr<ID3DXEffect>& effect,
			 const rde::ComPtr<ID3DXEffect>& wireEffect)
:	m_effect(effect),
	m_wireEffect(wireEffect)
{
}

Scene::~Scene()
{
	Clear();
}

void Scene::AddMesh(Mesh* m)
{
	RDE_ASSERT(m);
	m_meshes.push_back(m);
}

void Scene::AddEntity(Entity* ent)
{
	RDE_ASSERT(FindEntity(ent->m_id) == 0 && "Entity with specified ID already present");
	m_entities.insert(rde::make_pair(ent->m_id, ent));
	if (ent->m_mesh)
		AddMesh(ent->m_mesh);
}

void Scene::RemoveEntity(Entity* ent)
{
	Entities::iterator it = m_entities.find(ent->m_id);
	if (it != m_entities.end())
		m_entities.erase(it);
	for (Meshes::size_type i = 0; i < m_meshes.size(); ++i)
	{
		Mesh* mesh = m_meshes[i];
		if (mesh == ent->m_mesh)
		{
			m_meshes.erase(m_meshes.begin() + i);
			return;
		}
	}
}

void Scene::Clear()
{
	for (Meshes::iterator it = m_meshes.begin(); it != m_meshes.end(); ++it)
	{
		if ((*it)->m_parent == 0)
			delete *it;
	}

	for (Entities::iterator it = m_entities.begin(); it != m_entities.end(); ++it)
	{
		delete it->second;
	}
	m_entities.clear();
	m_meshes.clear();
}

Mesh* Scene::Pick(int sx, int sy)
{
	const D3DXMATRIX& projMatrix = m_camera.GetProjection();
	D3DXVECTOR3 v;
	const int width = m_camera.GetViewport().width;
	const int height = m_camera.GetViewport().height;
	v.x = (((2.f * sx) / width) - 1) / projMatrix._11;
	v.y = -(((2.f * sy) / height) - 1) / projMatrix._22;
	v.z = 1.f;
	const D3DXMATRIX& viewMatrix = m_camera.GetView();

	for (Meshes::iterator itMesh = m_meshes.begin(); itMesh != m_meshes.end(); ++itMesh)
	{
		// Ignore meshes with no entity parents.
		if ((*itMesh)->m_parent == 0)
			continue;

		const D3DXMATRIX& worldTransform = (*itMesh)->m_worldTransform;
		D3DXMATRIX worldView = worldTransform * viewMatrix;
		D3DXMATRIX viewToObject;
		::D3DXMatrixInverse(&viewToObject, 0, &worldView);
		D3DXVECTOR3 rayDir(	v.x * viewToObject._11 + v.y * viewToObject._21 + v.z * viewToObject._31,
							v.x * viewToObject._12 + v.y * viewToObject._22 + v.z * viewToObject._32,
							v.x * viewToObject._13 + v.y * viewToObject._23 + v.z * viewToObject._33);
		D3DXVECTOR3 rayOrigin(viewToObject._41, viewToObject._42, viewToObject._43);
		BOOL hit(false);
		DWORD face;
		float u, v, dist;
		::D3DXIntersect((*itMesh)->GetMesh(), &rayOrigin, &rayDir, &hit, &face, &u, &v, &dist, 0, 0);
		if (hit)
			return *itMesh;
	}
	return 0;
}

void Scene::Render()
{
	RDE_ASSERT(m_effect.Get());
	m_camera.UpdateTransforms();

	UINT numPasses(0);
	DX_CHECK(m_effect->Begin(&numPasses, 
		D3DXFX_DONOTSAVESTATE | D3DXFX_DONOTSAVESHADERSTATE)); 
	for (UINT pass = 0; pass < numPasses; ++pass)
	{
		DX_CHECK(m_effect->BeginPass(pass));

		for (Meshes::iterator itMesh = m_meshes.begin(); itMesh != m_meshes.end(); ++itMesh)
		{
			if ((*itMesh)->m_parent)
				(*itMesh)->m_parent->Update();

			const D3DXMATRIX& worldTransform = (*itMesh)->m_worldTransform;
			DX_CHECK(m_effect->SetMatrix("WorldMatrix", &worldTransform));
			D3DXMATRIX mvp;
			::D3DXMatrixMultiply(&mvp, &worldTransform, &m_camera.GetViewProjection());
			DX_CHECK(m_effect->SetMatrix("ModelViewProjection", &mvp));
			(*itMesh)->PreRender(m_effect.Get());
			DX_CHECK(m_effect->CommitChanges());
			(*itMesh)->Render();
		}
		DX_CHECK(m_effect->EndPass());
	}
	DX_CHECK(m_effect->End());
}

void Scene::RenderSelected(Mesh* selectedMesh)
{
	if (selectedMesh == 0)
		return;

	m_camera.UpdateTransforms();

	UINT numPasses(0);
	DX_CHECK(m_wireEffect->Begin(&numPasses, 
		D3DXFX_DONOTSAVESTATE | D3DXFX_DONOTSAVESHADERSTATE)); 
	for (UINT pass = 0; pass < numPasses; ++pass)
	{
		DX_CHECK(m_wireEffect->BeginPass(pass));

		const D3DXMATRIX& worldTransform = selectedMesh->m_worldTransform;
		D3DXMATRIX mvp;
		::D3DXMatrixMultiply(&mvp, &worldTransform, &m_camera.GetViewProjection());
		DX_CHECK(m_wireEffect->SetMatrix("ModelViewProjection", &mvp));
		DX_CHECK(m_wireEffect->CommitChanges());
		selectedMesh->Render();
		DX_CHECK(m_wireEffect->EndPass());
	}
	DX_CHECK(m_wireEffect->End());
}

Entity* Scene::FindEntity(rde::uint32_t id)
{
	Entities::iterator it = m_entities.find(id);
	return it == m_entities.end() ? 0 : it->second;
}
