#include "PreCompiled.h"
#include "Messages.h"
#include "ipc/Message.h"
#include "ipc/PipeConnection.h"
#include "Framework/Application.h"
#include "Framework/Entity.h"
#include "Framework/Mesh.h"
#include "Framework/Scene.h"
#include "core/Console.h"
#include "core/Thread.h"
#include <conio.h>

namespace
{
// Mouse movement can be either send by pipe from the client or
// retrieved using GetCursorPos(). The latter is obviously more
// effective, but it makes recording messages for later replay
// impossible.
#define RECEIVE_MOUSE_MOVE	1

namespace MouseButton
{
	enum Enum
	{
		LEFT,
		MIDDLE,
		RIGHT,
	};
}

struct Environment
{
	Environment(Application& a, Scene& s, rde::PipeConnection& pc)
	: app(a), scene(s), connection(pc), selectedMesh(0), draggedMesh(0) {}

	void InitScene();
	void OnMouseMove();
	void HandleMessage(const rde::Message&);
	void MessagePump();
	void HandleWindowMessage(UINT msg, WPARAM, LPARAM);
	void OnKeyDown(int key);
	void OnMouseDown(MouseButton::Enum);
	void OnMouseUp(MouseButton::Enum);

	void SelectEntity(rde::uint32_t id);

	Application& app;
	Scene& scene;
	rde::PipeConnection& connection;
	HWND hWnd;
	Mesh* selectedMesh;
	Mesh* draggedMesh;
	int mouseX, mouseY;
	float prevDragX, prevDragY;
};

void Environment::InitScene()
{
	Mesh* floor = app.CreateCube(1000.f, 0.1f, 1000.f);
	scene.AddMesh(floor);
}

D3DXVECTOR4 ARGBToVec4(rde::uint32_t argb)
{
	const rde::uint8_t a = (argb >> 24);
	const rde::uint8_t r = (argb >> 16) & 0xFF;
	const rde::uint8_t g = (argb >> 8) & 0xFF;
	const rde::uint8_t b = argb & 0xFF;
	return D3DXVECTOR4(float(r) / 255.f, float(g) / 255.f, float(b) / 255.f, 
		float(a) / 255.f);
}
rde::uint32_t Vec4ToARGB(const D3DXVECTOR4& v)
{
	const rde::uint32_t r = rde::uint32_t(v.x * 255.f);
	const rde::uint32_t g = rde::uint32_t(v.y * 255.f);
	const rde::uint32_t b = rde::uint32_t(v.z * 255.f);
	const rde::uint32_t a = rde::uint32_t(v.w * 255.f);
	return (a << 24) | (r << 16) | (g << 8) | b;
}

void Environment::HandleWindowMessage(UINT msg, WPARAM wp, LPARAM)
{
	switch (msg)
	{
	case WM_KEYDOWN:
		OnKeyDown((int)wp);
		break;
	case WM_LBUTTONDOWN:
		OnMouseDown(MouseButton::LEFT);
		break;
	case WM_LBUTTONUP:
		OnMouseUp(MouseButton::LEFT);
		break;
	}
}

void Environment::OnKeyDown(int key)
{
	if (key == 'W')
	{
		D3DXVECTOR3 camPos = scene.m_camera.GetPosition();
		camPos.z += 0.5f;
		scene.m_camera.SetPosition(camPos);
	}
	else if (key == 'S')
	{
		D3DXVECTOR3 camPos = scene.m_camera.GetPosition();
		camPos.z -= 0.5f;
		scene.m_camera.SetPosition(camPos);
	}
}

void Environment::SelectEntity(rde::uint32_t id)
{
	rde::Message* selMsg = connection.CreateMessage(MessageID::SELECT_OBJECT,
		sizeof(rde::uint32_t), rde::MessageType::USER);
	*(rde::uint32_t*)selMsg->GetData() = id;
	if (!connection.Send(selMsg))
		delete selMsg;
}

void Environment::OnMouseDown(MouseButton::Enum button)
{
	if (button == MouseButton::LEFT)
	{
		selectedMesh = scene.Pick(mouseX, mouseY);
		draggedMesh = selectedMesh;
		rde::uint32_t id = selectedMesh && selectedMesh->m_parent ? 
			selectedMesh->m_parent->m_id : Entity::kInvalidID;
		SelectEntity(id);

		const int width = scene.m_camera.GetViewport().width;
		const int height = scene.m_camera.GetViewport().height;
		const float x = float(mouseX - width/2) / width;
		const float y = -float(mouseY - height/2) / height;
		prevDragX = x;
		prevDragY = y;
	}
}
void Environment::OnMouseUp(MouseButton::Enum button)
{
	if (button == MouseButton::LEFT)
		draggedMesh = 0;
}

void Environment::HandleMessage(const rde::Message& msg)
{
	const rde::Message::Header& hdr = msg.GetHeader();
	switch (hdr.id)
	{
	case MessageID::CREATE_OBJECT:
		{
			Message_CreateObject* msgData = (Message_CreateObject*)msg.GetData();
			Mesh* mesh(0);
			const rde::uint8_t objType = msgData->objectType;
			if (objType == ObjectType::CUBE)
			{
				rde::Console::Printf("Creating cube\n");
				mesh = app.CreateCube(20.f, 20.f, 20.f);
			}
			else if (objType == ObjectType::SPHERE)
			{
				rde::Console::Printf("Creating sphere\n");
				mesh = app.CreateSphere(30.f);
			}
			else if (objType == ObjectType::TORUS)
			{
				rde::Console::Printf("Creating torus\n");
				mesh = app.CreateTorus(8.f, 12.f);
				*((int*)0) = 666;
			}
			else
			{
				rde::Console::Printf("Unknown mesh type [%d]\n", objType);
				RDE_ASSERT(!"Unknown mesh type");
			}
			mesh->ComputeBoundingBox(&app);
			mesh->m_color = D3DXVECTOR4(1.f, 0.f, 0.f, 1.f);
			Entity* ent = new Entity(msgData->id);
			ent->SetMesh(mesh);
			scene.AddEntity(ent);

			// Send return message letting know we succeeded.
			// (only if object not yet present on the client).
			if (msgData->id == Entity::kInvalidID)
			{
				rde::Message* retMsg = connection.CreateMessage(MessageID::CREATE_OBJECT,
					sizeof(Message_CreateObject), rde::MessageType::USER);
				msgData = (Message_CreateObject*)retMsg->GetData();
				msgData->id = ent->m_id;
				msgData->objectType = objType;
				if (!connection.Send(retMsg))
					delete retMsg;
			}
			// Select our new object.
			//SelectEntity(ent->m_id);
			//selectedMesh = mesh;
			break;
		}
	case MessageID::DELETE_OBJECT:
		{
			rde::uint32_t id = *(rde::uint32_t*)msg.GetData();
			Entity* ent = scene.FindEntity(id);
			if (ent != 0)
			{
				scene.RemoveEntity(ent);
				if (ent->m_mesh == selectedMesh)
					selectedMesh = 0;
				delete ent;

				rde::Message* retMsg = connection.CreateMessage(MessageID::DELETE_OBJECT,
					sizeof(rde::uint32_t), rde::MessageType::USER);
				*(rde::uint32_t*)retMsg->GetData() = id;
				if (!connection.Send(retMsg))
					delete retMsg;
			}
			break;
		}
	case MessageID::CHANGE_OBJECT_PROPERTIES:
		{
			Message_ObjectProperties* msgData = (Message_ObjectProperties*)msg.GetData();
			Entity* ent = scene.FindEntity(msgData->id);
			if (ent != 0)
			{
				ent->m_mesh->m_color = ARGBToVec4(msgData->color);
				ent->SetPosition(D3DXVECTOR3(msgData->x, msgData->y, msgData->z));
				ent->SetVitalityPoints(msgData->vitalityPoints);
			}
			else
			{
				rde::Console::Warningf("Trying to change properties of non-existing object (ID=0x%X)\n", 
					msgData->id);
			}
			break;
		}
	case MessageID::KEY_DOWN:
		{
			Message_KeyboardEvent* msgData = (Message_KeyboardEvent*)msg.GetData();
			HandleWindowMessage(WM_KEYDOWN, WPARAM(msgData->key), 0);
			break;
		}
	case MessageID::MOUSE_DOWN:
		{
			Message_MouseEvent* msgData = (Message_MouseEvent*)msg.GetData();
			const LPARAM lp = (mouseX) | (mouseY << 16);
			if (msgData->button == MouseButton::LEFT)
				HandleWindowMessage(WM_LBUTTONDOWN, 0, lp);
			else if (msgData->button == MouseButton::MIDDLE)
				HandleWindowMessage(WM_MBUTTONDOWN, 0, lp);
			else if (msgData->button == MouseButton::RIGHT)
				HandleWindowMessage(WM_RBUTTONDOWN, 0, lp);
			
			break;
		}
	case MessageID::MOUSE_UP:
		{
			Message_MouseEvent* msgData = (Message_MouseEvent*)msg.GetData();
			if (msgData->button == MouseButton::LEFT)
				HandleWindowMessage(WM_LBUTTONUP, 0, 0);
			else if (msgData->button == MouseButton::MIDDLE)
				HandleWindowMessage(WM_MBUTTONUP, 0, 0);
			else if (msgData->button == MouseButton::RIGHT)
				HandleWindowMessage(WM_RBUTTONUP, 0, 0);

			break;
		}
	case MessageID::MOUSE_MOVE:
		{
			Message_MouseMoveEvent* msgData = (Message_MouseMoveEvent*)msg.GetData();
			mouseX = msgData->x;
			mouseY = msgData->y;
			break;
		}
	case MessageID::VIEWPORT_RESIZE:
		{
			Message_Viewport* msgData = (Message_Viewport*)msg.GetData();
			scene.m_camera.SetViewport(0, 0, msgData->width, msgData->height);
			break;
		}
	case MessageID::CLEAR_WORLD:
		{
			scene.Clear();
			draggedMesh = selectedMesh = 0;
			InitScene();
			break;
		}
	case MessageID::CRASH_GAME:
		{
			*((int*)0) = 666;
			break;
		}
	default:
		RDE_ASSERT(!"Unknown message received");
	}
}

void Environment::MessagePump()
{
#if !RECEIVE_MOUSE_MOVE
	POINT ptCursor;
	::GetCursorPos(&ptCursor);
	::ScreenToClient(hWnd, &ptCursor);
	mouseX = ptCursor.x;
	mouseY = ptCursor.y;
#endif
	rde::Message* msg(0);
	while (connection.Receive(&msg))
	{
		HandleMessage(*msg);
		delete msg;
	}
	OnMouseMove();
}

void Environment::OnMouseMove()
{
	static const float kDragScale = 200.f;
	if (draggedMesh != 0)
	{
		const int width = scene.m_camera.GetViewport().width;
		const int height = scene.m_camera.GetViewport().height;
		const float x = float(mouseX - width/2) / width;
		const float y = -float(mouseY - height/2) / height;
		const float dx = (x - prevDragX) * kDragScale;
		const float dy = (y - prevDragY) * kDragScale;

		D3DXVECTOR3 pos = draggedMesh->m_parent->GetPosition();
		pos.x += dx;
		pos.y += dy;
		draggedMesh->m_parent->SetPosition(pos);

		// Do not send update if no real movement.
		if (fabsf(dx) > 1e-4f || fabsf(dy) > 1e-4f)
		{
			rde::Message* msg = connection.CreateMessage(MessageID::CHANGE_OBJECT_PROPERTIES, 
				sizeof(Message_ObjectProperties), rde::MessageType::USER);
			Message_ObjectProperties* msgData = (Message_ObjectProperties*)msg->GetData();
			msgData->id = draggedMesh->m_parent->m_id;
			msgData->x = pos.x;
			msgData->y = pos.y;
			msgData->z = pos.z;
			msgData->color = Vec4ToARGB(draggedMesh->m_color);
			if (!connection.Send(msg))
				delete msg;
		}
		prevDragX = x;
		prevDragY = y;
	}
}

}


int main()
{
	rde::PipeConnection connection;
	if (!connection.ServerInit("IPCTest", "IPCTestPipe"))
		return 1;

	// Wait for connection.
	while (connection.GetState() != rde::ConnectionState::ACTIVE)
		rde::Thread::Sleep(100);

	rde::Message* msg(0);
	connection.Receive(&msg, true);
	if (!msg)
		return 1;

	Message_Handshake* msgData = (Message_Handshake*)msg->GetData();
	HWND hWnd = (HWND)msgData->hwnd;
	const int width = msgData->width;
	const int height = msgData->height;
	Entity::SetNextID(msgData->nextID);
	delete msg;

	msg = connection.CreateMessage(MessageID::HANDSHAKE, sizeof(Message_Handshake),
		rde::MessageType::USER);
	connection.Send(msg);


	{
		Application app;
		app.Init(hWnd, width, height);
		Scene scene(app.CreateEffect("global.fx"), app.CreateEffect("wire.fx"));
		::Environment env(app, scene, connection);
		env.InitScene();

		scene.m_camera.SetViewport(0, 0, width, height);
		scene.m_camera.SetPosition(D3DXVECTOR3(0.f, 5.f, -100.f));
		env.hWnd = hWnd;
		while (!kbhit())
		{
			env.MessagePump();
			app.Frame(0.f, scene, env.selectedMesh);
		}
	}
	return 0;
}
