#include "ipc/Pipe.h"
#include "core/Console.h"
#include "core/RdeAssert.h"
#include "core/ThreadEvent.h"
#include "core/win32/Windows.h"

namespace rde
{
struct Pipe::Impl
{
	Impl():	handle(0), connected(false) 
	{
	    memset(&overlapped, 0, sizeof(overlapped));
		overlapped.hEvent = ::CreateEvent(0, true, false, 0);
	}
	~Impl()
	{
		CloseHandle(overlapped.hEvent);
	}

	HANDLE		handle;
	OVERLAPPED	overlapped;
	bool		connected;
};

Pipe::Pipe()
{
	m_impl = new Impl();
}
Pipe::~Pipe()
{
	Close();
	delete m_impl;
	m_impl = 0;
}

bool Pipe::Create(const char* name)
{
	m_impl->handle = ::CreateNamedPipe(name, 
                                     PIPE_ACCESS_DUPLEX | FILE_FLAG_OVERLAPPED,
                                     PIPE_TYPE_BYTE | PIPE_READMODE_BYTE | PIPE_WAIT,
                                     1,
                                     kBufferSize,
                                     kBufferSize,
                                     1000,
                                     NULL);
	if (m_impl->handle == INVALID_HANDLE_VALUE)
	{
		rde::Console::Warningf("Cannot create pipe '%s'\n", name);
		return false;
	}
	return true;
}

bool Pipe::Open(const char* name)
{
	if (!::WaitNamedPipe(name, 0))
		return false;

	m_impl->handle = ::CreateFile(name, 
                                GENERIC_READ | GENERIC_WRITE, 
                                0,
                                NULL,
                                OPEN_EXISTING,
                                FILE_FLAG_OVERLAPPED,
                                NULL);
	if (m_impl->handle == INVALID_HANDLE_VALUE)
	{
		char err[512];
		Sys::GetErrorString(err, RDE_ARRAY_COUNT(err));
		Console::Printf("Pipe Support: Failed to open pipe (%s)\n", err);
		return false;
	}
    DWORD mode = PIPE_READMODE_BYTE | PIPE_WAIT; 
    if (!::SetNamedPipeHandleState(m_impl->handle, &mode, NULL, NULL))
    {
		char err[512];
		Sys::GetErrorString(err, RDE_ARRAY_COUNT(err));
		Console::Printf("Pipe Support: Failed to set client byte mode (%s)\n", err);
		::CloseHandle(m_impl->handle);
		return false;
    }
	return true;
}

bool Pipe::Connect(const ThreadEvent& evtTerminate)
{
	OVERLAPPED connect;
	memset(&connect, 0, sizeof(connect));
	connect.hEvent = ::CreateEvent(0, true, false, 0);

	if (!::ConnectNamedPipe(m_impl->handle, &connect))
	{
		const DWORD error = ::GetLastError();
		if (error == ERROR_PIPE_CONNECTED)
		{
			::CloseHandle(connect.hEvent);
			return true;
		}
		if (error != ERROR_IO_PENDING)
		{
			char err[512];
			Sys::GetErrorString(err, RDE_ARRAY_COUNT(err));
			Console::Warningf("Failed to connect pipe (%s)\n", err);
			::CloseHandle(connect.hEvent);
			return false;
		}
		else
		{
			HANDLE events[] = { evtTerminate.GetSystemRepresentation(), connect.hEvent };
			const DWORD result = ::WaitForMultipleObjects(2, events, FALSE, INFINITE);

			RDE_ASSERT(result != WAIT_FAILED);
			if (result)
			{
				// Terminated?
				if ((result - WAIT_OBJECT_0) == 0)
				{
					::CloseHandle(connect.hEvent);
					::CancelIo(m_impl->handle);
					return false;
				}
			}
		}

		DWORD bytes;
		if (!::GetOverlappedResult(m_impl->handle, &connect, &bytes, false))
		{
			char err[512];
			Sys::GetErrorString(err, RDE_ARRAY_COUNT(err));
			Console::Warningf("Failed to connect pipe (%s)\n", err);
	        ::CloseHandle(connect.hEvent);
		    return false;
		}
	}

	::CloseHandle(connect.hEvent);
	return true;
}

bool Pipe::Read(void* buffer, uint32_t bytes, const ThreadEvent& evtTerminate, 
		uint32_t& bytesRead)
{
	if (bytes == 0)
		return true;

	DWORD read(0);
#if RDE_IPC_DEBUG_LEVEL > 0
	char err[512];
#endif
	if (!::ReadFile(m_impl->handle, buffer, bytes, &read, &m_impl->overlapped))
	{
		if (::GetLastError() != ERROR_IO_PENDING)
		{
#if RDE_IPC_DEBUG_LEVEL > 0
			Sys::GetErrorString(err, RDE_ARRAY_COUNT(err));
			Console::Warningf("Failed to initiate Pipe::Read (%s)\n", err);
#endif
			return false;
		}
		else
		{
			HANDLE events[] = { evtTerminate.GetSystemRepresentation(), 
				m_impl->overlapped.hEvent };
			const DWORD result = ::WaitForMultipleObjects(2, events, FALSE, INFINITE);
			RDE_ASSERT(result != WAIT_FAILED);
			if (result == WAIT_OBJECT_0)
			{
				::CancelIo(m_impl->handle);
				return false;
			}
			if (!::GetOverlappedResult(m_impl->handle, &m_impl->overlapped, &read, false))
			{
#if RDE_IPC_DEBUG_LEVEL > 0
				Sys::GetErrorString(err, RDE_ARRAY_COUNT(err));
				Console::Warningf("Pipe::Read failed (%s)\n", err);
#endif
				return false;
			}
		}
	}
	bytesRead = read;
	return true;
}

bool Pipe::Write(const void* buffer, uint32_t bytes, const ThreadEvent& evtTerminate, 
		uint32_t& bytesWritten)
{
	if (bytes == 0)
		return true;

#if RDE_IPC_DEBUG_LEVEL > 0
	char err[512];
#endif
	DWORD written(0);
	if (!::WriteFile(m_impl->handle, buffer, bytes, &written, &m_impl->overlapped))
	{
		if (::GetLastError() != ERROR_IO_PENDING)
		{
#if RDE_IPC_DEBUG_LEVEL > 0
			Sys::GetErrorString(err, RDE_ARRAY_COUNT(err));
			Console::Warningf("Failed to initiate Pipe::Write (%s)\n", err);
#endif
			return false;
		}
		else
		{
			HANDLE events[] = { evtTerminate.GetSystemRepresentation(), 
				m_impl->overlapped.hEvent };
			const DWORD result = ::WaitForMultipleObjects(2, events, FALSE, INFINITE);
			RDE_ASSERT(result != WAIT_FAILED);
			if (result == WAIT_OBJECT_0)
			{
				::CancelIo(m_impl->handle);
				return false;
			}
			if (!::GetOverlappedResult(m_impl->handle, &m_impl->overlapped, &written, false))
			{
#if RDE_IPC_DEBUG_LEVEL > 0
				Sys::GetErrorString(err, RDE_ARRAY_COUNT(err));
				Console::Warningf("Pipe::Write failed (%s)\n", err);
#endif
				return false;
			}
		}
	}
	bytesWritten = written;
	return true;
}

void Pipe::Disconnect()
{
	RDE_ASSERT(m_impl->connected);
	if (!::FlushFileBuffers(m_impl->handle))
	{
#if RDE_IPC_DEBUG_LEVEL > 0
		char err[512];
		Sys::GetErrorString(err, RDE_ARRAY_COUNT(err));
		Console::Printf("Pipe Support: Failed to flush pipe buffers (%s)\n", err);
#endif
	}
	if (!::DisconnectNamedPipe(m_impl->handle))
	{
#if RDE_IPC_DEBUG_LEVEL > 0
		char err[512];
		Sys::GetErrorString(err, RDE_ARRAY_COUNT(err));
		Console::Printf("Pipe Support: Failed to diconnect pipe (%s)\n", err);
#endif
	}
	m_impl->connected = false;
}

void Pipe::Close()
{
	if (m_impl->connected)
		Disconnect();
	if (m_impl->handle != 0)
		::CloseHandle(m_impl->handle);
	m_impl->handle = 0;
}

}
